//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

// Lighting constants.
// Now why not 32 levels here?
#define LIGHTLEVELS		16
#define LIGHTSEGSCALE	0.0625

#define MAXLIGHTZ		128

// Number of diminishing brightness levels.
// There a 0-31, i.e. 32 LUT in the COLORMAP lump.
#define NUMCOLORMAPS	32

//	Theoretically cliping can give only 4 new vertexes. In practice due to
// roundof errors we can get more extra vertexes
#define NUM_EXTRA_VERTS	16
#define MAX_STACK_VERTS	64

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

float					centerxfrac;
float					centeryfrac;
float					xprojection;
float					yprojection;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static draw_vertex_t	*p_vert;
static int				r_emited;
static int				max_emited;

//
// Lighting LUT.
// Used for z-depth cuing per column/row,
//  and other lighting effects (sector ambient, flash).
//
static float			zlight[LIGHTLEVELS][MAXLIGHTZ];

static TCvarF			d_light_z_scale("d_light_z_scale", "0.0625");
static TCvarI			d_use_zlight("d_use_zlight", "1", CVAR_ARCHIVE);

// CODE --------------------------------------------------------------------

//==========================================================================
//
//  TSoftwareDrawer::InitLightTables
//
//==========================================================================

void TSoftwareDrawer::InitLightTables(void)
{
	int			i;
	int			j;
	float		light_level;
	float		startmap;

    //	Calculate the light levels to use for each
    // light_level / distance combination.
    for (i = 0; i < LIGHTLEVELS; i++)
    {
        startmap = ((float)LIGHTLEVELS - 1.0 - (float)i) * (2.0 * 256.0 / (float)LIGHTLEVELS);
        for (j = 0; j < MAXLIGHTZ; j++)
        {
            light_level = startmap - 640.0 / ((float)j + 1.0);

            if (light_level < 0)
                light_level = 0;

            if (light_level > 255)
                light_level = 255;

            zlight[i][j] = 255 - light_level;
        }
    }
}

//==========================================================================
//
//	TransformVector
//
//==========================================================================

void TransformVector(const TVec &in, TVec &out)
{
	out.x = DotProduct(in, viewright);
	out.y = DotProduct(in, viewup);
	out.z = DotProduct(in, viewforward);
}

//==========================================================================
//
//	ProjectVertex
//
//==========================================================================

static void ProjectVertex(draw_vertex_t &v)
{
	float z1 = 1.0f / v.z;
	v.x = v.x * z1 * xprojection + centerxfrac;
	v.y = v.y * z1 * yprojection + centeryfrac;

	if (v.x < -0.5)
		v.x = -0.5;
	if (v.x > viewwidth - 0.5)
		v.x = viewwidth - 0.5;
	if (v.y < -0.5)
		v.y = -0.5;
	if (v.y > viewheight - 0.5)
		v.y = viewheight - 0.5;
}

//==========================================================================
//
//	GetLight
//
//==========================================================================

static float GetLight(float l, float dist)
{
	float light;
	if (fixedlight)
	{
		light = fixedlight;
	}
	else if (d_use_zlight)
	{
//		int lightlevel = (int)(l * LIGHTSEGSCALE + extralight);
		int lightlevel = (int)(l * LIGHTSEGSCALE);
		if (lightlevel >= LIGHTLEVELS)
			lightlevel = LIGHTLEVELS - 1;
		if (lightlevel < 0)
			lightlevel = 0;

		int index = (int)(dist * d_light_z_scale);
		if (index >= MAXLIGHTZ)
			index = MAXLIGHTZ - 1;

		light = zlight[lightlevel][index];
	}
	else
	{
//		light = l + extralight * 16.0;
		light = l;
		if (light > 255)
			light = 255;
	}
	return 255 - light;
}

//==========================================================================
//
//	EmitVertex
//
//==========================================================================

static void EmitVertex(const draw_vertex_t &v)
{
	if (r_emited >= max_emited)
		return;

	draw_vertex_t	tr;

	TransformVector(v - vieworg, tr);
	if (tr.z < 0.01)
		tr.z = 0.01;

	tr.l = GetLight(v.l, Length(tr));

	ProjectVertex(tr);

	p_vert[r_emited] = tr;
	r_emited++;
}

//==========================================================================
//
//	ClipEdge
//
//==========================================================================

static void ClipEdge(const draw_vertex_t &v0, const draw_vertex_t &v1,
	TClipPlane *clip, int clipflags)
{
	if (clip)
	{
		do
		{
			if (!(clipflags & clip->clipflag))
			{
				continue;
			}

			float d0 = DotProduct(v0, clip->normal) - clip->dist;
			float d1 = DotProduct(v1, clip->normal) - clip->dist;
			if (d0 >= 0)
			{
				// point 0 is unclipped
				if (d1 >= 0)
				{
					// both points are unclipped
					continue;
				}

				// only point 1 is clipped

				draw_vertex_t &clipvert = clip->enter;
				clip->entered = true;

				float f = d0 / (d0 - d1);
				clipvert.x = v0.x + f * (v1.x - v0.x);
				clipvert.y = v0.y + f * (v1.y - v0.y);
				clipvert.z = v0.z + f * (v1.z - v0.z);
				clipvert.l = v0.l + f * (v1.l - v0.l);

				ClipEdge(v0, clipvert, clip->next, clipflags);
				if (clip->exited)
				{
					clip->entered = false;
					clip->exited = false;
					ClipEdge(clipvert, clip->exit, view_clipplanes, clipflags ^ clip->clipflag);
				}
				return;
			}
			else
			{
				// point 0 is clipped

				if (d1 < 0)
				{
					// both points are clipped
					return;
				}

				// only point 0 is clipped

				draw_vertex_t &clipvert = clip->exit;
				clip->exited = true;

				float f = d0 / (d0 - d1);
				clipvert.x = v0.x + f * (v1.x - v0.x);
				clipvert.y = v0.y + f * (v1.y - v0.y);
				clipvert.z = v0.z + f * (v1.z - v0.z);
				clipvert.l = v0.l + f * (v1.l - v0.l);

				if (clip->entered)
				{
					clip->entered = false;
					clip->exited = false;
					ClipEdge(clip->enter, clipvert, view_clipplanes, clipflags ^ clip->clipflag);
				}
				ClipEdge(clipvert, v1, clip->next, clipflags);
				return;
			}
		} while ((clip = clip->next) != NULL);
	}

	//	add the vertex
	EmitVertex(v1);
}

//==========================================================================
//
//	TSoftwareDrawer::DrawPolygon
//
//==========================================================================

void TSoftwareDrawer::DrawPolygon(draw_vertex_t *cv, int count, int texture,
	int clipflags, bool use_fog)
{
	int		i;

	for (i = 0; i < 4; i++)
	{
		view_clipplanes[i].entered = false;
		view_clipplanes[i].exited = false;
	}

#ifdef __GNUC__
	max_emited = count + NUM_EXTRA_VERTS;
	draw_vertex_t dv[max_emited];
#else
	max_emited = MAX_STACK_VERTS;
	draw_vertex_t dv[MAX_STACK_VERTS];
#endif
	r_emited = 0;
	p_vert = dv;

	for (i = 0; i < count; i++)
	{
		ClipEdge(cv[i ? i - 1 : count - 1], cv[i], view_clipplanes, clipflags);
	}

	if (r_emited < 3)
	{
		return;
	}

	if (texture & TEXF_FLAT)
	{
		SetFlat(texture);
	}
	else
	{
		SetTexture(texture);
	}

	spanfunc = D_DrawSpan;

	if (fixedcolormap)
	{
		ds_colormap = fixedcolormap;
		ds_colormap16 = fixedcolormap16;
	}
	else if (use_fog)
	{
		ds_colormap = fogmap;
		ds_colormap16 = fogmap16;
	}
	else
	{
		ds_colormap = fadetable;
		ds_colormap16 = fadetable16;
	}

	D_DrawWorldPolygon(dv, r_emited, false);
}

//==========================================================================
//
//	TSoftwareDrawer::DrawMaskedPolygon
//
//==========================================================================

void TSoftwareDrawer::DrawMaskedPolygon(draw_vertex_t *cv, int count, int texture,
	int translucency, int clipflags, bool use_fog)
{
	int		i;

	for (i = 0; i < 4; i++)
	{
		view_clipplanes[i].entered = false;
		view_clipplanes[i].exited = false;
	}

#ifdef __GNUC__
	max_emited = count + NUM_EXTRA_VERTS;
	draw_vertex_t dv[max_emited];
#else
	max_emited = MAX_STACK_VERTS;
	draw_vertex_t dv[MAX_STACK_VERTS];
#endif
	r_emited = 0;
	p_vert = dv;

	for (i = 0; i < count; i++)
	{
		ClipEdge(cv[i ? i - 1 : count - 1], cv[i], view_clipplanes, clipflags);
	}

	if (r_emited < 3)
	{
		return;
	}

	if (texture & TEXF_FLAT)
	{
		SetFlat(texture);
	}
	else
	{
		SetTexture(texture);
	}

	if (translucency < 5)
	{
		spanfunc = D_DrawMaskedSpan;
	}
	else
	{
		int trindex = (translucency - 5) / 10;
		if (trindex < 5)
		{
			ds_transluc = tinttables[trindex];
			spanfunc = D_DrawFuzzSpan;
		}
		else
		{
			ds_transluc = tinttables[8 - trindex];
			spanfunc = D_DrawAltFuzzSpan;
		}
		ds_transluc16 = translucency * 256 / 100;
	}

	if (fixedcolormap)
	{
		ds_colormap = fixedcolormap;
		ds_colormap16 = fixedcolormap16;
	}
	else if (use_fog)
	{
		ds_colormap = fogmap;
		ds_colormap16 = fogmap16;
	}
	else
	{
		ds_colormap = fadetable;
		ds_colormap16 = fadetable16;
	}

	if (!translucency)
	{
		D_DrawWorldPolygon(dv, r_emited, true);
	}
	else
	{
		D_DrawTranslucentWorldPolygon(dv, r_emited);
	}
}

//==========================================================================
//
//	TSoftwareDrawer::DrawZPolygon
//
//==========================================================================

void TSoftwareDrawer::DrawZPolygon(draw_vertex_t *cv, int count, int clipflags)
{
	int		i;

	for (i = 0; i < 4; i++)
	{
		view_clipplanes[i].entered = false;
		view_clipplanes[i].exited = false;
	}

#ifdef __GNUC__
	max_emited = count + NUM_EXTRA_VERTS;
	draw_vertex_t dv[max_emited];
#else
	max_emited = MAX_STACK_VERTS;
	draw_vertex_t dv[MAX_STACK_VERTS];
#endif
	r_emited = 0;
	p_vert = dv;

	for (i = 0; i < count; i++)
	{
		ClipEdge(cv[i ? i - 1 : count - 1], cv[i], view_clipplanes, clipflags);
	}

	if (r_emited < 3)
	{
		return;
	}

	spanfunc = D_DrawZSpan;

	D_DrawWorldPolygon(dv, r_emited, false);
}

//==========================================================================
//
//	TSoftwareDrawer::DrawSkyPolygon
//
//==========================================================================

void TSoftwareDrawer::DrawSkyPolygon(draw_vertex_t *cv, int count,
	int texture, bool double_sky)
{
	int		i;

#ifdef __GNUC__
	draw_vertex_t dv[count];
#else
	draw_vertex_t dv[MAX_STACK_VERTS];
	if (count > MAX_STACK_VERTS) count = MAX_STACK_VERTS;
#endif

	for (i = 0; i < count; i++)
	{
		draw_vertex_t &v = cv[i];
		draw_vertex_t &tr = dv[i];

		TransformVector(v, tr);
		tr.l = v.l;
		if (tr.z < 0.01)
			tr.z = 0.01;
		ProjectVertex(tr);
	}

	SetSkyTexture(texture, double_sky);

	spanfunc = double_sky ? D_DrawDoubleSkySpan : D_DrawSkySpan;
	ds_colormap = fadetable;
	ds_colormap16 = fadetable16;

	r_texorg += vieworg;
	r_dist += DotProduct(vieworg, r_normal);
	D_DrawWorldPolygon(dv, count, true);
	r_dist -= DotProduct(vieworg, r_normal);
	r_texorg -= vieworg;
}

//==========================================================================
//
//	TSoftwareDrawer::DrawSpritePolygon
//
//==========================================================================

void TSoftwareDrawer::DrawSpritePolygon(draw_vertex_t *cv, int lump,
	int translucency, int translation, bool use_fog)
{
	int		i;

	for (i = 0; i < 4; i++)
	{
		view_clipplanes[i].entered = false;
		view_clipplanes[i].exited = false;
	}

	draw_vertex_t dv[16];
	p_vert = dv;
	max_emited = 16;
	r_emited = 0;

	for (i = 0; i < 4; i++)
	{
		ClipEdge(cv[i ? i - 1 : 3], cv[i], view_clipplanes, 15);
	}

	if (r_emited < 3)
	{
		return;
	}

	if (translucency < 5)
	{
		spritespanfunc = D_DrawSpriteSpan;
	}
	else
	{
		int trindex = (translucency - 5) / 10;
		if (trindex < 5)
		{
			ds_transluc = tinttables[trindex];
			spritespanfunc = D_DrawFuzzSpriteSpan;
		}
		else
		{
			ds_transluc = tinttables[8 - trindex];
			spritespanfunc = D_DrawAltFuzzSpriteSpan;
		}
		ds_transluc16 = translucency * 256 / 100;
	}

	if (fixedcolormap)
	{
		ds_colormap = fixedcolormap;
		ds_colormap16 = fixedcolormap16;
	}
	else if (use_fog)
	{
		ds_colormap = fogmap;
		ds_colormap16 = fogmap16;
	}
	else
	{
		ds_colormap = fadetable;
		ds_colormap16 = fadetable16;
	}

	TVec center((cv[0].x + cv[1].x + cv[2].x + cv[3].x) * 0.25,
		(cv[0].y + cv[1].y + cv[2].y + cv[3].y) * 0.25,
		(cv[0].z + cv[1].z + cv[2].z + cv[3].z) * 0.25);
	int cmapnum = (int)GetLight(cv[0].l, Length(center - vieworg)) / 8;
	SetSpriteLump(lump, cmapnum, translation);

	// -1 (-epsilon) so we never wander off the edge of the texture
	bbextents = (spritewidth[lump] << 16) - 1;
	bbextentt = (spriteheight[lump] << 16) - 1;

	D_SpriteDrawPolygon(dv, r_emited);
}

