//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

#define MAX_POLYS			2000
#define MAX_EDGES			8000
#define MAX_SPANS			MAXSCREENWIDTH
#define MAX_MASKED_SPANS	1000

// TYPES -------------------------------------------------------------------

struct EdgeT
{
	EdgeT	*next;
	int		lines;	/* number of lines to be sampled on this edge */
	fixed_t	u;
	fixed_t	u_step;
	fixed_t	l;
	fixed_t	lstepv;
};

struct DrawPoly_t
{
	DrawPoly_t	*prev;
	DrawPoly_t	*next;

	EdgeT		*eLeft;
	EdgeT		*eRight;

	float		zistepu;
	float		zistepv;
	float		ziorigin;

	float		sdivzstepu;
	float		sdivzstepv;
	float		tdivzstepu;
	float		tdivzstepv;
	float		sdivzorigin;
	float		tdivzorigin;
	fixed_t		sadjust;
	fixed_t		tadjust;

	byte*		colormap;
	word*		colormap16;
	byte*		source;
	spanfunc_t	spanfunc;
	int			masked;
};

struct span_t
{
	span_t		*next;
	DrawPoly_t	*poly;
	int			u;
	int			count;
	fixed_t		l;
	fixed_t		l_step;
};

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

#ifndef USEASM
int				ds_u;
int				ds_v;
int				ds_count;
float			d_zistepu;
float			d_zistepv;
float			d_ziorigin;
float			d_sdivzstepu;
float			d_tdivzstepu;
float			d_sdivzstepv;
float			d_tdivzstepv;
float			d_sdivzorigin;
float			d_tdivzorigin;
fixed_t			sadjust;
fixed_t			tadjust;
fixed_t			bbextents;
fixed_t			bbextentt;
fixed_t			ds_l;
fixed_t			ds_lstepu;

byte*			ds_source;
byte*			ds_colormap;
word*			ds_colormap16;
int				ds_tex_width;
byte*			ds_transluc;// For translucent spans
int				ds_transluc16;
#endif

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static DrawPoly_t		dpolys[MAX_POLYS];
static EdgeT			dedges[MAX_EDGES];
static int				numpolys;
static int				numedges;

static DrawPoly_t		polylist;
static DrawPoly_t		*polystoadd[MAXSCREENHEIGHT];

static span_t			dspans[MAX_SPANS];
static int				numspans;
static span_t			*spanlist;
static span_t			dmaskedspans[MAX_MASKED_SPANS];
static int				nummaskedspans;
static span_t			*maskedspanlist;

static int				outofpolys;
static int				outofedges;

static int				minindex;
static int				maxindex;

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	D_ClearPolys
//
//==========================================================================

void D_ClearPolys(void)
{
	numpolys = 0;
	numedges = 0;
	polylist.next = NULL;
	memset(polystoadd, 0, sizeof(polystoadd));
	outofpolys = 0;
	outofedges = 0;
}

//==========================================================================
//
//	AddPoly
//
//==========================================================================

static void AddPoly(DrawPoly_t *poly)
{
	DrawPoly_t *prev = &polylist;
	while (prev->next && prev->next < poly)
	{
		prev = prev->next;
	}
	poly->prev = prev;
	poly->next = prev->next;
	if (prev->next)
	{
		prev->next->prev = poly;
	}
	prev->next = poly;
}

//==========================================================================
//
//	RemovePoly
//
//==========================================================================

static void	RemovePoly(DrawPoly_t *poly)
{
	if (poly->next)
	{
		poly->next->prev = poly->prev;
	}
	poly->prev->next = poly->next;
}

//==========================================================================
//
//	D_MakeSpans
//
//==========================================================================

void D_MakeSpans(DrawPoly_t *poly)
{
	// initialize the span interpolants to the leftmost value
	int u0 = poly->eLeft->u >> 16;
	int u1 = poly->eRight->u >> 16;
	if (u0 >= u1)
	{
		return;
	}
	int count = u1 - u0;

	int l_step = 0;
	if (count > 1)
	{
		l_step = (poly->eRight->l - poly->eLeft->l) / (count - 1);
	}

	span_t *prev;
	span_t *next;
	span_t		*s;

	prev = NULL;
	next = spanlist;

	if (!next || next->u >= u1)
	{
		//	Empty list or first span starts after this span
		s = &dspans[numspans++];

		s->u = u0;
		s->count = count;
		s->l = poly->eLeft->l;
		s->l_step = l_step;
		s->poly = poly;
		s->next = next;
		spanlist = s;
		return;
	}

	if (next->u > u0)
	{
		//	Emit a fragment at the beginning of the list
		s = &dspans[numspans++];

		s->u = u0;
		s->count = next->u - u0;
		s->l = poly->eLeft->l;
		s->l_step = l_step;
		s->poly = poly;
		s->next = next;
		spanlist = s;

		prev = next;
		next = next->next;
		goto emit_fragments;
	}

	do
	{
		prev = next;
		next = next->next;
	}
	while (next && next->u <= u0);

	if (prev->u + prev->count > u0)
	{
		goto emit_fragments;
	}

	if (!next || next->u >= u1)
	{
		//	Completely fits between two spans
		s = &dspans[numspans++];

		s->u = u0;
		s->count = count;
		s->l = poly->eLeft->l;
		s->l_step = l_step;
		s->poly = poly;
		s->next = next;
		prev->next = s;
		return;
	}

	if (next->u > u0)
	{
		//	Emit a fragment with clipped end
		s = &dspans[numspans++];

		s->u = u0;
		s->count = next->u - u0;
		s->l = poly->eLeft->l;
		s->l_step = l_step;
		s->poly = poly;
		s->next = next;
		prev->next = s;

		prev = next;
		next = next->next;
	}

 emit_fragments:

	int startu;

	startu = prev->u + prev->count;
	if (startu >= u1)
	{
		//	End is clipped
		return;
	}

	if (!next || next->u >= u1)
	{
		//	Emit end
		s = &dspans[numspans++];

		s->u = startu;
		s->count = u1 - startu;
		s->l = poly->eLeft->l + (startu - u0) * l_step;
		s->l_step = l_step;
		s->poly = poly;
		s->next = next;
		prev->next = s;
		return;
	}

	if (next->u > startu)
	{
		//	Emit a fragment
		s = &dspans[numspans++];

		s->u = startu;
		s->count = next->u - startu;
		s->l = poly->eLeft->l + (startu - u0) * l_step;
		s->l_step = l_step;
		s->poly = poly;
		s->next = next;
		prev->next = s;
	}

	prev = next;
	next = next->next;

	goto emit_fragments;
}

//==========================================================================
//
//	D_MakeMaskedSpans
//
//==========================================================================

void D_MakeMaskedSpans(DrawPoly_t *poly)
{
	// initialize the span interpolants to the leftmost value
	int u0 = poly->eLeft->u >> 16;
	int u1 = poly->eRight->u >> 16;
	if (u0 >= u1)
	{
		return;
	}
	int count = u1 - u0;

	int l_step = 0;
	if (count > 1)
	{
		l_step = (poly->eRight->l - poly->eLeft->l) / (count - 1);
	}

	span_t *prev;
	span_t *next;
	span_t		*s;

	prev = NULL;
	next = spanlist;

	if (!next || next->u >= u1)
	{
		//	Empty list or first span starts after this span
		if (nummaskedspans >= MAX_MASKED_SPANS)
		{
			return;
		}
		s = &dmaskedspans[nummaskedspans++];

		s->u = u0;
		s->count = count;
		s->l = poly->eLeft->l;
		s->l_step = l_step;
		s->poly = poly;
		s->next = maskedspanlist;
		maskedspanlist = s;
		return;
	}

	if (next->u > u0)
	{
		//	Emit a fragment at the beginning of the list
		if (nummaskedspans >= MAX_MASKED_SPANS)
		{
			return;
		}
		s = &dmaskedspans[nummaskedspans++];

		s->u = u0;
		s->count = next->u - u0;
		s->l = poly->eLeft->l;
		s->l_step = l_step;
		s->poly = poly;
		s->next = maskedspanlist;
		maskedspanlist = s;

		prev = next;
		next = next->next;
		goto emit_fragments;
	}

	do
	{
		prev = next;
		next = next->next;
	}
	while (next && next->u <= u0);

	if (prev->u + prev->count > u0)
	{
		goto emit_fragments;
	}

	if (!next || next->u >= u1)
	{
		//	Completely fits between two spans
		if (nummaskedspans >= MAX_MASKED_SPANS)
		{
			return;
		}
		s = &dmaskedspans[nummaskedspans++];

		s->u = u0;
		s->count = count;
		s->l = poly->eLeft->l;
		s->l_step = l_step;
		s->poly = poly;
		s->next = maskedspanlist;
		maskedspanlist = s;
		return;
	}

	if (next->u > u0)
	{
		//	Emit a fragment with clipped end
		if (nummaskedspans >= MAX_MASKED_SPANS)
		{
			return;
		}
		s = &dmaskedspans[nummaskedspans++];

		s->u = u0;
		s->count = next->u - u0;
		s->l = poly->eLeft->l;
		s->l_step = l_step;
		s->poly = poly;
		s->next = maskedspanlist;
		maskedspanlist = s;

		prev = next;
		next = next->next;
	}

 emit_fragments:

	int startu;

	startu = prev->u + prev->count;
	if (startu >= u1)
	{
		//	End is clipped
		return;
	}

	if (!next || next->u >= u1)
	{
		//	Emit end
		if (nummaskedspans >= MAX_MASKED_SPANS)
		{
			return;
		}
		s = &dmaskedspans[nummaskedspans++];

		s->u = startu;
		s->count = u1 - startu;
		s->l = poly->eLeft->l + (startu - u0) * l_step;
		s->l_step = l_step;
		s->poly = poly;
		s->next = maskedspanlist;
		maskedspanlist = s;
		return;
	}

	if (next->u > startu)
	{
		//	Emit a fragment
		if (nummaskedspans >= MAX_MASKED_SPANS)
		{
			return;
		}
		s = &dmaskedspans[nummaskedspans++];

		s->u = startu;
		s->count = next->u - startu;
		s->l = poly->eLeft->l + (startu - u0) * l_step;
		s->l_step = l_step;
		s->poly = poly;
		s->next = maskedspanlist;
		maskedspanlist = s;
	}

	prev = next;
	next = next->next;

	goto emit_fragments;
}

//==========================================================================
//
//	D_AdvanceEdge
//
//==========================================================================

EdgeT *D_AdvanceEdge(EdgeT *edge)
{
	//	Advance to the next scan line.
	edge->lines--;
	if (!edge->lines)
	{
		return edge->next;
	}
	edge->u += edge->u_step;
	edge->l += edge->lstepv;
	return edge;
}

//==========================================================================
//
//	D_DrawPolyLine
//
//==========================================================================

void D_DrawPolyLine(DrawPoly_t *poly)
{
	if (poly->masked)
	{
		D_MakeMaskedSpans(poly);
	}
	else
	{
		D_MakeSpans(poly);
	}

	poly->eLeft = D_AdvanceEdge(poly->eLeft);
	poly->eRight = D_AdvanceEdge(poly->eRight);

	if (!poly->eLeft || !poly->eRight)
	{
		Z_ChangeTag(poly->source, PU_CACHE);
		RemovePoly(poly);
	}
}

//==========================================================================
//
//	D_EmitEdge
//
//==========================================================================

EdgeT *D_EmitEdge(draw_vertex_t *v0, draw_vertex_t *v1)
{
	int ceilv0 = (int)ceil(v0->y);
	int ceilv1 = (int)ceil(v1->y);

	if (ceilv1 <= ceilv0)
	{
		return NULL;
	}

	EdgeT *edge;

	edge = &dedges[numedges];
	numedges++;

	edge->lines = ceilv1 - ceilv0;

	float u_step = (v1->x - v0->x) / (v1->y - v0->y);
	float u = v0->x + ((float)ceilv0 - v0->y) * u_step;

	edge->u = (fixed_t)(u * 0x10000) + 0xffff;
	edge->u_step = (fixed_t)(u_step * 0x10000);

	edge->l = (int)v0->l << FRACBITS;
	if (edge->lines > 1)
	{
		edge->lstepv = (((int)v1->l << FRACBITS) - edge->l) / (edge->lines - 1);
	}
	return edge;
}

//==========================================================================
//
//	D_AddLeftEdges
//
//==========================================================================

EdgeT* D_AddLeftEdges(draw_vertex_t *vb, int count)
{
	EdgeT *list;
	int i;

	list = NULL;
	i = maxindex;
	do
	{
		int other;

		other = i - 1;
		if (other < 0)
		{
			other = count - 1;
		}
		EdgeT *edge = D_EmitEdge(&vb[other], &vb[i]);
		if (edge)
		{
			edge->next = list;
			list = edge;
		}
		i = other;
	} while (i != maxindex);
	return list;
}

//==========================================================================
//
//	D_AddRightEdges
//
//==========================================================================

EdgeT* D_AddRightEdges(draw_vertex_t *vb, int count)
{
	EdgeT *list;
	int i;

	list = NULL;
	i = maxindex;
	do
	{
		int other;

		other = i + 1;
		if (other >= count)
		{
			other = 0;
		}
		EdgeT *edge = D_EmitEdge(&vb[other], &vb[i]);
		if (edge)
		{
			edge->next = list;
			list = edge;
		}
		i = other;
	} while (i != maxindex);
	return list;
}

//==========================================================================
//
//	D_CalcGradients
//
//==========================================================================

void D_CalcGradients(DrawPoly_t *poly)
{
	TVec p_normal;

	TransformVector(r_normal, p_normal);
	float distinv = 1.0 / (r_dist - DotProduct(vieworg, r_normal));

	poly->zistepu = p_normal.x * distinv / xprojection;
	poly->zistepv = p_normal.y * distinv / yprojection;
	poly->ziorigin = p_normal.z * distinv -
			centerxfrac * poly->zistepu -
			centeryfrac * poly->zistepv;

	TVec tr_saxis;
	TVec tr_taxis;

	TransformVector(r_saxis, tr_saxis);
	TransformVector(r_taxis, tr_taxis);

	poly->sdivzstepu = tr_saxis.x / xprojection;
	poly->sdivzstepv = tr_saxis.y / yprojection;

	poly->tdivzstepu = tr_taxis.x / xprojection;
	poly->tdivzstepv = tr_taxis.y / yprojection;

	poly->sdivzorigin = tr_saxis.z - centerxfrac * poly->sdivzstepu - centeryfrac * poly->sdivzstepv;
	poly->tdivzorigin = tr_taxis.z - centerxfrac * poly->tdivzstepu - centeryfrac * poly->tdivzstepv;

	poly->sadjust = (fixed_t)(DotProduct(vieworg - r_texorg, r_saxis) * 0x10000);
	poly->tadjust = (fixed_t)(DotProduct(vieworg - r_texorg, r_taxis) * 0x10000);
}

//==========================================================================
//
//	D_DrawWorldPolygon
//
//==========================================================================

void D_DrawWorldPolygon(draw_vertex_t *verts, int count, int masked)
{
	float ymin, ymax;
	int i;

	//	Find the top and bottom vertices, and make sure there's at least one
	// scan to draw
	ymin = 999999.9;
	ymax = -999999.9;

	for (i = 0; i < count; i++)
	{
		if (verts[i].y < ymin)
		{
			ymin = verts[i].y;
			minindex = i;
		}

		if (verts[i].y > ymax)
		{
			ymax = verts[i].y;
			maxindex = i;
		}
	}

	ymin = ceil(ymin);
	ymax = ceil(ymax);

	if (ymin >= ymax)
	{
		// doesn't cross any scans at all
		return;
	}

	if (numpolys >= MAX_POLYS)
	{
		outofpolys++;
		return;
	}

	if (numedges + count > MAX_EDGES)
	{
		outofedges += count;
		return;
	}

	DrawPoly_t *poly = &dpolys[numpolys];

	D_CalcGradients(poly);

	poly->colormap = ds_colormap;
	poly->colormap16 = ds_colormap16;
	poly->source = ds_source;
	poly->spanfunc = spanfunc;
	poly->masked = masked;

	Z_ChangeTag(ds_source, PU_STATIC);

	poly->eLeft = D_AddLeftEdges(verts, count);
	poly->eRight = D_AddRightEdges(verts, count);

	if (!poly->eLeft || !poly->eRight)
	{
		//	No edges??? Should never happen.
		return;
	}

	poly->next = polystoadd[(int)ymin];
	polystoadd[(int)ymin] = poly;

	numpolys++;
}

//==========================================================================
//
//	D_DrawSpans
//
//==========================================================================

static void D_DrawSpans(span_t *s)
{
	while (s)
	{
		ds_u = s->u;
		ds_count = s->count;
		ds_l = s->l;
		ds_lstepu = s->l_step;

		d_zistepu = s->poly->zistepu;
		d_zistepv = s->poly->zistepv;
		d_ziorigin = s->poly->ziorigin;
		d_sdivzstepu = s->poly->sdivzstepu;
		d_tdivzstepu = s->poly->tdivzstepu;
		d_sdivzstepv = s->poly->sdivzstepv;
		d_tdivzstepv = s->poly->tdivzstepv;
		d_sdivzorigin = s->poly->sdivzorigin;
		d_tdivzorigin = s->poly->tdivzorigin;
		sadjust = s->poly->sadjust;
		tadjust = s->poly->tadjust;

		ds_colormap = s->poly->colormap;
		ds_colormap16 = s->poly->colormap16;
		ds_source = s->poly->source;

		s->poly->spanfunc();
		s = s->next;
	}
}

//==========================================================================
//
//	TSoftwareDrawer::WorldDrawing
//
//==========================================================================

void TSoftwareDrawer::WorldDrawing(void)
{
	DrawPoly_t *poly;

	for (ds_v = 0; ds_v < viewheight; ds_v++)
	{
		numspans = 0;
		spanlist = NULL;
		nummaskedspans = 0;
		maskedspanlist = NULL;

		poly = polystoadd[ds_v];
		while (poly)
		{
			DrawPoly_t *next;
			next = poly->next;
			AddPoly(poly);
			poly = next;
		}

		poly = polylist.next;
		while (poly)
		{
			DrawPoly_t *next = poly->next;
			D_DrawPolyLine(poly);
			poly = next;
		}

		D_DrawSpans(spanlist);
		D_DrawSpans(maskedspanlist);
	}
	if (outofpolys) cond << "Out of " << outofpolys << " polys\n";
	if (outofedges) cond << "Out of " << outofedges << " edges\n";
}

//==========================================================================
//
//	D_DrawTranslucentWorldPolygon
//
//==========================================================================

void D_DrawTranslucentWorldPolygon(draw_vertex_t *verts, int count)
{
	float ymin, ymax;
	int i;

	//	Find the top and bottom vertices, and make sure there's at least one
	// scan to draw
	ymin = 999999.9;
	ymax = -999999.9;

	for (i = 0; i < count; i++)
	{
		if (verts[i].y < ymin)
		{
			ymin = verts[i].y;
			minindex = i;
		}

		if (verts[i].y > ymax)
		{
			ymax = verts[i].y;
			maxindex = i;
		}
	}

	ymin = ceil(ymin);
	ymax = ceil(ymax);

	if (ymin >= ymax)
	{
		// doesn't cross any scans at all
		return;
	}

	TVec p_normal;

	TransformVector(r_normal, p_normal);
	float distinv = 1.0 / (r_dist - DotProduct(vieworg, r_normal));

	d_zistepu = p_normal.x * distinv / xprojection;
	d_zistepv = p_normal.y * distinv / yprojection;
	d_ziorigin = p_normal.z * distinv -
			centerxfrac * d_zistepu -
			centeryfrac * d_zistepv;

	TVec tr_saxis;
	TVec tr_taxis;

	TransformVector(r_saxis, tr_saxis);
	TransformVector(r_taxis, tr_taxis);

	d_sdivzstepu = tr_saxis.x / xprojection;
	d_sdivzstepv = tr_saxis.y / yprojection;

	d_tdivzstepu = tr_taxis.x / xprojection;
	d_tdivzstepv = tr_taxis.y / yprojection;

	d_sdivzorigin = tr_saxis.z - centerxfrac * d_sdivzstepu - centeryfrac * d_sdivzstepv;
	d_tdivzorigin = tr_taxis.z - centerxfrac * d_tdivzstepu - centeryfrac * d_tdivzstepv;

	sadjust = (fixed_t)(DotProduct(vieworg - r_texorg, r_saxis) * 0x10000);
	tadjust = (fixed_t)(DotProduct(vieworg - r_texorg, r_taxis) * 0x10000);

	numedges = 0;

	EdgeT *eLeft = D_AddLeftEdges(verts, count);
	EdgeT *eRight = D_AddRightEdges(verts, count);

	for (ds_v = (int)ymin; eLeft && eRight && ds_v < viewheight; ds_v++)
	{
		// initialize the span interpolants to the leftmost value
		int u0 = eLeft->u >> 16;
		int u1 = eRight->u >> 16;
		int count = u1 - u0;
		if (count > 0)
		{
			ds_u = u0;
			ds_count = count;
			ds_l = eLeft->l;
			if (count > 1)
			{
				ds_lstepu = (eRight->l - eLeft->l) / (count - 1);
			}
			spanfunc();
		}
		eLeft = D_AdvanceEdge(eLeft);
		eRight = D_AdvanceEdge(eRight);
	}
}


