//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

#define SPRITE_CACHE_SIZE			256

// TYPES -------------------------------------------------------------------

struct sprite_cache_t
{
	void*		data;
	void*		cmap;
	int			lump;
	int			tnum;
};

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static byte**		texturedata;
static byte**		skytexturedata;	//FIXME this is a waste of memory
static byte**		flatdata;

static sprite_cache_t	sprite_cache[SPRITE_CACHE_SIZE];
static int				sprite_cache_count;

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	TSoftwareDrawer::InitTextures
//
//==========================================================================

void TSoftwareDrawer::InitTextures(void)
{
	//	Textures
    texturedata = (byte**)Z_Calloc(numtextures * 4);
    skytexturedata = (byte**)Z_Calloc(numtextures * 4);
	//	Flats
    flatdata = (byte**)Z_Calloc((numflats) * 4, PU_STATIC, 0);
}

//==========================================================================
//
//	D_FlushCaches
//
//==========================================================================

void D_FlushCaches(void)
{
	int		i;

	for (i = 0; i < SPRITE_CACHE_SIZE; i++)
	{
		if (sprite_cache[i].data)
		{
			Z_Free(sprite_cache[i].data);
		}
	}
}

//==========================================================================
//
//	DrawColumnInCache
//
// 	Clip and draw a column from a patch into a flat buffer.
//
//		column - column to draw
//		cache - buffer
//		originx, originy - position of column in the buffer
//		cachewidth, cacheheight - size of the cache
//
//==========================================================================

static void DrawColumnInCache(column_t* column, byte* cache,
	int originx, int originy, int cachewidth, int cacheheight)
{
    int		count;
    int		position;
    byte*	source;
    byte*	dest;
	
	// step through the posts in a column
    while (column->topdelta != 0xff)
    {
		source = (byte *)column + 3;
		count = column->length;
		position = originy + column->topdelta;

		//	Clip position
		if (position < 0)
		{
	    	count += position;
			source -= position;
	    	position = 0;
		}
		if (position + count > cacheheight)
		{
	    	count = cacheheight - position;
		}
    	dest = cache + originx + position * cachewidth;

    	while (count-- > 0)
    	{
			*dest = *source ? *source : black_color;
			source++;
			dest += cachewidth;
    	}
		
		column = (column_t *)((byte *)column + column->length + 4);
    }
}

//==========================================================================
//
//	GenerateTexture
//
// 	Using the texture definition, the composite texture is created from the
// patches, and each column is cached.
//
//==========================================================================

static void GenerateTexture(int texnum)
{
    byte*			block;
    texdef_t*		texture;
    texpatch_t*		patch;	
    patch_t*		realpatch;
    int				x;
    int				x1;
    int				x2;
    int				i;
    column_t*		patchcol;
	int				wtimes;
	int				htimes;
	int				texture_width;

    texture = textures[texnum];

	texture_width = texture->width;
	if (texture_width > 256)
	{
		cond << "JL in trouble - texture " << texture->name
			<< " haves width " << texture_width << " greater than 256\n";
		texture_width = 256;
	}

    block = (byte*)Z_Calloc(256 * 256, PU_STATIC, (void**)&texturedata[texnum]);

    // Composite the columns together.
    patch = texture->patches;

	wtimes = 256 / texture_width;
	htimes = 256 / texture->height;

    for (i = 0; i < texture->patchcount; i++, patch++)
    {
		realpatch = (patch_t*)W_CacheLumpNum(patch->patch, PU_TEMP);
		x1 = patch->originx;
		x2 = x1 + LittleShort(realpatch->width);

		if (x1 < 0)
	    	x = 0;
		else
	    	x = x1;
	
		if (x2 > texture_width)
	    	x2 = texture_width;

		for ( ; x < x2; x++)
		{
	    	patchcol = (column_t *)((byte *)realpatch
				    + LittleLong(realpatch->columnofs[x - x1]));
			for (int ht = 0; ht < htimes; ht++)
			{
				for (int wt = 0; wt < wtimes; wt++)
				{
			    	DrawColumnInCache(patchcol, block + wt * texture_width +
			    		ht * texture->height * 256, x, patch->originy,
			    		256, texture->height);
				}
			}
		}
    }

    // Now that the texture has been built in column cache,
    //  it is purgable from zone memory.
    Z_ChangeTag(block, PU_CACHE);
}

//==========================================================================
//
// 	TSoftwareDrawer::SetTexture
//
//==========================================================================

void TSoftwareDrawer::SetTexture(int tex)
{
	tex = texturetranslation[tex];

    if (!texturedata[tex])
		GenerateTexture(tex);

    ds_source = texturedata[tex];
}

//==========================================================================
//
//	GenerateSkyTexture
//
//==========================================================================

static void GenerateSkyTexture(int texnum)
{
    byte*			block;
    texdef_t*		texture;
    texpatch_t*		patch;	
    patch_t*		realpatch;
    int				x;
    int				x1;
    int				x2;
    int				i;
    column_t*		patchcol;
	int				wtimes;
	int				htimes;

    texture = textures[texnum];

    block = (byte*)Z_Calloc(1024 * 256, PU_STATIC, (void**)&skytexturedata[texnum]);

    // Composite the columns together.
    patch = texture->patches;

	wtimes = 1024 / texture->width;
	htimes = 256 / texture->height;

    for (i = 0; i < texture->patchcount; i++, patch++)
    {
		realpatch = (patch_t*)W_CacheLumpNum(patch->patch, PU_TEMP);
		x1 = patch->originx;
		x2 = x1 + LittleShort(realpatch->width);

		if (x1 < 0)
	    	x = 0;
		else
	    	x = x1;
	
		if (x2 > texture->width)
	    	x2 = texture->width;

		for ( ; x < x2; x++)
		{
	    	patchcol = (column_t *)((byte *)realpatch
				    + LittleLong(realpatch->columnofs[x - x1]));
			for (int ht = 0; ht < htimes; ht++)
			{
				for (int wt = 0; wt < wtimes; wt++)
				{
			    	DrawColumnInCache(patchcol, block + wt * texture->width +
			    		ht * texture->height * 1024, x, patch->originy,
			    		1024, texture->height);
				}
			}
		}
    }

    // Now that the texture has been built in column cache,
    //  it is purgable from zone memory.
    Z_ChangeTag(block, PU_CACHE);
}

//==========================================================================
//
// 	TSoftwareDrawer::SetSkyTexture
//
//==========================================================================

void TSoftwareDrawer::SetSkyTexture(int tex, bool)
{
	tex = texturetranslation[tex];

    if (!skytexturedata[tex])
		GenerateSkyTexture(tex);

    ds_source = skytexturedata[tex];
}

//==========================================================================
//
//	TSoftwareDrawer::SetFlat
//
//==========================================================================

void TSoftwareDrawer::SetFlat(int num)
{
	num &= ~TEXF_FLAT;
	num = flattranslation[num];

	if (!flatdata[num])
	{
		byte *block = (byte*)Z_Malloc(256 * 256, PU_STATIC, (void**)&flatdata[num]);
		byte *data = (byte*)W_CacheLumpNum(flatlumps[num], PU_TEMP);
		for (int j = 0; j < 256; j++)
		{
			for (int i = 0; i < 4; i++)
			{
				memcpy(block + (j << 8) + (i << 6), data + ((j & 63) << 6), 64);
			}
		}
		Z_ChangeTag(block, PU_CACHE);
	}

    ds_source = flatdata[num];
}

//==========================================================================
//
//	GenerateSprite
//
//==========================================================================

static void	GenerateSprite(int lump, int slot, void *cmap, int translation)
{
    patch_t	*patch = (patch_t*)W_CacheLumpNum(spritelumps[lump], PU_STATIC);

	int w = LittleShort(patch->width);
	int h = LittleShort(patch->height);

    void *block = (byte*)Z_Calloc(w * h * PixelBytes, PU_CACHE,
    	&sprite_cache[slot].data);
	sprite_cache[slot].cmap = cmap;
	sprite_cache[slot].lump = lump;
	sprite_cache[slot].tnum = translation;

	byte *trtab = translationtables + translation * 256;

	for (int x = 0; x < w; x++)
	{
    	column_t *column = (column_t *)((byte *)patch + LittleLong(patch->columnofs[x]));

		// step through the posts in a column
	    while (column->topdelta != 0xff)
	    {
		    byte* source = (byte *)column + 3;
			int count = column->length;

			if (ScreenBPP == 8)
			{
			    byte* dest = ((byte*)block) + x + column->topdelta * w;
		    	while (count--)
	    		{
					*dest = ((byte*)cmap)[trtab[*source]];
					source++;
					dest += w;
		    	}
			}
			else
			{
			    word* dest = ((word*)block) + x + column->topdelta * w;
		    	while (count--)
	    		{
					*dest = ((word*)cmap)[trtab[*source]];
					source++;
					dest += w;
		    	}
			}
			column = (column_t *)((byte *)column + column->length + 4);
	    }
	}

	Z_ChangeTag(patch, PU_CACHE);
}

//==========================================================================
//
//	TSoftwareDrawer::SetSpriteLump
//
//==========================================================================

void TSoftwareDrawer::SetSpriteLump(int lump, int cmapnum, int translation)
{
	void *cmap;
	if (ScreenBPP == 8)
	{
		cmap = ds_colormap + cmapnum * 256;
	}
	else
	{
		cmap = ds_colormap16 + cmapnum * 256;
	}

	int i;
	int avail = -1;
	for (i = 0; i <	SPRITE_CACHE_SIZE; i++)
	{
		if (sprite_cache[i].data)
		{
			if (sprite_cache[i].lump == lump &&
				sprite_cache[i].cmap == cmap &&
				sprite_cache[i].tnum == translation)
			{
				ds_source = (byte*)sprite_cache[i].data;
				ds_tex_width = spritewidth[lump];
				return;
			}
		}
		else
		{
			if (avail < 0)
				avail = i;
		}
	}
	if (avail < 0)
	{
		Z_Free(sprite_cache[sprite_cache_count].data);
		avail = sprite_cache_count;
		sprite_cache_count = (sprite_cache_count + 1) % SPRITE_CACHE_SIZE;
	}

	GenerateSprite(lump, avail, cmap, translation);
	ds_source = (byte*)sprite_cache[avail].data;
	ds_tex_width = spritewidth[lump];
}

