//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**	
//**	Refresh of things, i.e. objects represented by sprites.
//**
//** 	Sprite rotation 0 is facing the viewer, rotation 1 is one angle turn
//**  CLOCKWISE around the axis. This is not the same as the angle, which
//**  increases counter clockwise (protractor). There was a lot of stuff
//**  grabbed wrong, so I changed it...
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "gamedefs.h"
#include "r_local.h"

// MACROS ------------------------------------------------------------------

#define MAX_SPRITE_MODELS	512
#define MAX_TRANS_SPRITES	256

// TYPES -------------------------------------------------------------------

enum
{
	SPR_VP_PARALLEL_UPRIGHT,
	SPR_FACING_UPRIGHT,
	SPR_VP_PARALLEL,
	SPR_ORIENTED,
	SPR_VP_PARALLEL_ORIENTED,
	SPR_VP_PARALLEL_UPRIGHT_ORIENTED,
};

//
//  Sprites are patches with a special naming convention
// so they can be recognized by R_InitSprites.
//  The base name is NNNNFx or NNNNFxFx, with
// x indicating the rotation, x = 0, 1-7.
//  The sprite and frame specified by a thing_t
// is range checked at run time.
//  A sprite is a patch_t that is assumed to represent
// a three dimensional object and may have multiple
// rotations pre drawn.
//  Horizontal flipping is used to save space,
// thus NNNNF2F5 defines a mirrored patch.
//  Some sprites will only have one picture used
// for all views: NNNNF0
//
struct spriteframe_t
{
	// If false use 0 for any position.
	// Note: as eight entries are available,
	//  we might as well insert the same name eight times.
	boolean     rotate;

    // Lump to use for view angles 0-7.
	short		lump[8];

	// Flip bit (1 = flip) to use for view angles 0-7.
	bool		flip[8];
};

//
// 	A sprite definition:
// a number of animation frames.
//
struct spritedef_t
{
	int					numframes;
	spriteframe_t		*spriteframes;
};

struct trans_sprite_t
{
	draw_vertex_t	*dv;
	int				count;
	int				lump;
	TVec			normal;
	float			pdist;
	TVec			saxis;
	TVec			taxis;
	TVec			texorg;
	int				translucency;
	int				translation;
	bool			use_fog;
	bool			is_sprite;
	float			dist;
};

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

//	Temporary variables for sprite installing
static spriteframe_t	sprtemp[30];
static int				maxframe;
static const char		*spritename;

//	variables used to look up
// and range check thing_t sprites patches
static spritedef_t		sprites[MAX_SPRITE_MODELS];

static float			sprlight;

static TCvarI			r_draw_mobjs("r_draw_mobjs", "1", CVAR_ARCHIVE);
static TCvarI			r_draw_psprites("r_draw_psprites", "1", CVAR_ARCHIVE);

static TCvarI			croshair("croshair", "0", CVAR_ARCHIVE);
static TCvarI			croshair_trans("croshair_trans", "0", CVAR_ARCHIVE);

static trans_sprite_t	trans_sprites[MAX_TRANS_SPRITES];
static draw_vertex_t	trans_sprite_verts[4 * MAX_TRANS_SPRITES];

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	InstallSpriteLump
//
//	Local function for R_InitSprites.
//
//==========================================================================

static void InstallSpriteLump(int lumpnr, int frame, int rotation, bool flipped)
{
	int			r;

	if ((dword)frame >= 30 || (dword)rotation > 8)
	{
		I_Error("InstallSpriteLump: "
				"Bad frame characters in lump %i", spritelumps[lumpnr]);
	}

	if (frame > maxframe)
		maxframe = frame;

	if (rotation == 0)
	{
		// the lump should be used for all rotations
		sprtemp[frame].rotate = false;
		for (r = 0; r < 8; r++)
		{
			sprtemp[frame].lump[r] = lumpnr;
			sprtemp[frame].flip[r] = flipped;
		}
     	return;
	}

	// the lump is only used for one rotation
	if (sprtemp[frame].rotate == false)
	{
		for (r = 0; r < 8; r++)
		{
			sprtemp[frame].lump[r] = -1;
			sprtemp[frame].flip[r] = false;
		}
	}

	sprtemp[frame].rotate = true;

	// make 0 based
	rotation--;

	sprtemp[frame].lump[rotation] = lumpnr;
	sprtemp[frame].flip[rotation] = flipped;
}

//==========================================================================
//
//	R_InstallSprite
//
// 	Builds the sprite rotation matrixes to account for horizontally flipped
// sprites. Will report an error if the lumps are inconsistant.
// 	Sprite lump names are 4 characters for the actor, a letter for the frame,
// and a number for the rotation. A sprite that is flippable will have an
// additional letter/number appended. The rotation character can be 0 to
// signify no rotations.
//
//==========================================================================

void R_InstallSprite(const char *name, int index)
{
	int			intname;
	int			l;
	int			frame;
	int			rotation;

	if ((dword)index >= MAX_SPRITE_MODELS)
	{
		Host_Error("Invalid sprite index %d for sprite %s", index, name);
	}
	spritename = name;
	memset(sprtemp, -1, sizeof(sprtemp));
	maxframe = -1;

    // scan all the lump names for each of the names,
    //  noting the highest frame letter.
    // Just compare 4 characters as ints
	intname = *(int *)spritename;

	// scan the lumps, filling in the frames for whatever is found
	for (l = 0; l < numspritelumps; l++)
	{
		const char *lumpname;
		lumpname = W_LumpName(spritelumps[l]);
		if (*(int *)lumpname == intname)
		{
			frame = lumpname[4] - 'A';
			rotation = lumpname[5] - '0';
			InstallSpriteLump(l, frame, rotation, false);

			if (lumpname[6])
			{
				frame = lumpname[6] - 'A';
				rotation = lumpname[7] - '0';
				InstallSpriteLump(l, frame, rotation, true);
			}
		}
	}

	// check the frames that were found for completeness
	if (maxframe == -1)
	{
		sprites[index].numframes = 0;
		return;
	}

	maxframe++;

	for (frame = 0 ; frame < maxframe ; frame++)
	{
		switch ((int)sprtemp[frame].rotate)
		{
		 case -1:
			// no rotations were found for that frame at all
			I_Error("R_InstallSprite: No patches found "
					"for %s frame %c", spritename, frame + 'A');
			break;

		 case 0:
			// only the first rotation is needed
			break;

		 case 1:
			// must have all 8 frames
			for (rotation = 0; rotation < 8; rotation++)
			{
				if (sprtemp[frame].lump[rotation] == -1)
				{
					I_Error("R_InstallSprite: Sprite %s frame %c "
							"is missing rotations", spritename, frame + 'A');
				}
			}
			break;
		}
	}

	if (sprites[index].spriteframes)
	{
		Z_Free(sprites[index].spriteframes);
		sprites[index].spriteframes = NULL;
	}
	// allocate space for the frames present and copy sprtemp to it
	sprites[index].numframes = maxframe;
	sprites[index].spriteframes = (spriteframe_t*)
		Z_StrMalloc(maxframe * sizeof(spriteframe_t));
	memcpy(sprites[index].spriteframes, sprtemp, maxframe * sizeof(spriteframe_t));
}

//==========================================================================
//
//	GetSpriteParams
//
//==========================================================================

static void	GetSpriteParams(int lump)
{
    patch_t	*patch = (patch_t*)W_CacheLumpNum(spritelumps[lump], PU_TEMP);

	spritewidth[lump] = LittleShort(patch->width);
	spriteheight[lump] = LittleShort(patch->height);
	spriteoffset[lump] = LittleShort(patch->leftoffset);
	spritetopoffset[lump] = LittleShort(patch->topoffset);
}

//==========================================================================
//
//	R_AddTranslucentPoly
//
//==========================================================================

void R_DrawTranslucentPoly(draw_vertex_t *sv, int count, float dist, int lump,
	int translucency, int translation, bool is_sprite)
{
	int found = -1;
	float best_dist = -1;
	for (int i = 0; i < MAX_TRANS_SPRITES; i++)
	{
		trans_sprite_t &spr = trans_sprites[i];
		if (!spr.translucency)
		{
			if (count <= 4)
				spr.dv = trans_sprite_verts + 4 * i;
			else
				spr.dv = (draw_vertex_t*)Z_StrMalloc(sizeof(draw_vertex_t) * count);
			memcpy(spr.dv, sv, sizeof(draw_vertex_t) * count);
			spr.count = count;
			spr.dist = dist;
			spr.lump = lump;
			spr.normal = r_normal;
			spr.pdist = r_dist;
			spr.saxis = r_saxis;
			spr.taxis = r_taxis;
			spr.texorg = r_texorg;
			spr.translucency = translucency;
			spr.translation = translation;
			spr.use_fog = r_use_fog;
			spr.is_sprite = is_sprite;
			return;
		}
		if (spr.dist > best_dist)
		{
			found = i;
			best_dist = spr.dist;
		}
	}
	if (best_dist > dist)
	{
		//	All slots are full, draw and replace a far away sprite
		TVec tmp_normal = r_normal;
		float tmp_dist = r_dist;
		TVec tmp_saxis = r_saxis;
		TVec tmp_taxis = r_taxis;
		TVec tmp_texorg = r_texorg;
		trans_sprite_t &spr = trans_sprites[found];
		r_normal = spr.normal;
		r_dist = spr.pdist;
		r_saxis = spr.saxis;
		r_taxis = spr.taxis;
		r_texorg = spr.texorg;
		if (spr.is_sprite)
		{
			Drawer->DrawSpritePolygon(spr.dv, spr.lump,
				spr.translucency, spr.translation, spr.use_fog);
		}
		else
		{
			Drawer->DrawMaskedPolygon(spr.dv, spr.count, spr.lump,
				spr.translucency, 15, spr.use_fog);
			if (spr.count > 4)
				Z_Free(spr.dv);
		}
		if (count <= 4)
			spr.dv = trans_sprite_verts + 4 * found;
		else
			spr.dv = (draw_vertex_t*)Z_StrMalloc(sizeof(draw_vertex_t) * count);
		memcpy(spr.dv, sv, sizeof(draw_vertex_t) * count);
		spr.count = count;
		spr.dist = dist;
		spr.lump = lump;
		spr.normal = tmp_normal;
		spr.pdist = tmp_dist;
		spr.saxis = tmp_saxis;
		spr.taxis = tmp_taxis;
		spr.texorg = tmp_texorg;
		spr.translucency = translucency;
		spr.translation = translation;
		spr.use_fog = r_use_fog;
		return;
	}

	//	All slots are full and are nearer to current sprite so draw it
	if (is_sprite)
	{
		Drawer->DrawSpritePolygon(sv, lump, translucency,
			translation, r_use_fog);
	}
	else
	{
		Drawer->DrawMaskedPolygon(sv, count, lump, translucency, 15, r_use_fog);
	}
}

//==========================================================================
//
//	RenderSprite
//
//==========================================================================

extern TCvarI		r_chasecam;

static void RenderSprite(mobj_base_t *thing)
{
	if (thing == &cl_mobjs[cl.origin_id] && !r_chasecam)
	{
		//	Don't draw client's mobj
		return;
	}

	if (thing->translucency >= 95)
	{
		// Never make a vissprite when MF2_DONTDRAW is flagged.
		return;
	}

	int spr_type = thing->spritetype;

	TVec sprorigin = thing->origin;
	TVec sprforward;
	TVec sprright;
	TVec sprup;

	float		dot;
	TVec		tvec;
	float		sr;
	float		cr;

	switch (spr_type)
	{
	 case SPR_VP_PARALLEL_UPRIGHT:
		//	Generate the sprite's axes, with sprup straight up in worldspace,
		// and sprright parallel to the viewplane. This will not work if the
		// view direction is very close to straight up or down, because the
		// cross product will be between two nearly parallel vectors and
		// starts to approach an undefined state, so we don't draw if the two
		// vectors are less than 1 degree apart
		dot = viewforward.z;	//	same as DotProduct(viewforward, sprup)
								// because sprup is 0, 0, 1
		if ((dot > 0.999848) || (dot < -0.999848))	// cos(1 degree) = 0.999848
			return;

		sprup = TVec(0, 0, 1);
		//	CrossProduct(sprup, viewforward)
		sprright = Normalize(TVec(viewforward.y, -viewforward.x, 0));
		//	CrossProduct(sprright, sprup)
		sprforward = TVec(-sprright.y, sprright.x, 0);
		break;

	 case SPR_FACING_UPRIGHT:
		//	Generate the sprite's axes, with sprup straight up in worldspace,
		// and sprright perpendicular to sprorigin. This will not work if the
		// view direction is very close to straight up or down, because the
		// cross product will be between two nearly parallel vectors and
		// starts to approach an undefined state, so we don't draw if the two
		// vectors are less than 1 degree apart
		tvec = Normalize(sprorigin - vieworg);
		dot = tvec.z;	//	same as DotProduct (tvec, sprup) because
						// sprup is 0, 0, 1
		if ((dot > 0.999848) || (dot < -0.999848))	// cos(1 degree) = 0.999848
			return;
		sprup = TVec(0, 0, 1);
		//	CrossProduct(sprup, -sprorigin)
		sprright = Normalize(TVec(tvec.y, -tvec.x, 0));
		//	CrossProduct(sprright, sprup)
		sprforward = TVec(-sprright.y, sprright.x, 0);
		break;

	 case SPR_VP_PARALLEL:
		//	Generate the sprite's axes, completely parallel to the viewplane.
		// There are no problem situations, because the sprite is always in
		// the same position relative to the viewer
		sprup = viewup;
		sprright = viewright;
		sprforward = viewforward;
		break;

	 case SPR_ORIENTED:
		//	Generate the sprite's axes, according to the sprite's world
		// orientation
		AngleVectors(thing->angles, sprforward, sprright, sprup);
		break;

	 case SPR_VP_PARALLEL_ORIENTED:
		//	Generate the sprite's axes, parallel to the viewplane, but
		// rotated in that plane around the center according to the sprite
		// entity's roll angle. So sprforward stays the same, but sprright
		// and sprup rotate
		sr = msin(thing->angles.roll);
		cr = mcos(thing->angles.roll);

		sprforward = viewforward;
		sprright = TVec(viewright.x * cr + viewup.x * sr, viewright.y * cr +
			viewup.y * sr, viewright.z * cr + viewup.z * sr);
		sprup = TVec(viewright.x * -sr + viewup.x * cr, viewright.y * -sr +
			viewup.y * cr, viewright.z * -sr + viewup.z * cr);
		break;

	 case SPR_VP_PARALLEL_UPRIGHT_ORIENTED:
		//	Generate the sprite's axes, with sprup straight up in worldspace,
		// and sprright parallel to the viewplane and then rotated in that
		// plane around the center according to the sprite entity's roll
		// angle. So sprforward stays the same, but sprright and sprup rotate
		// This will not work if the view direction is very close to straight
		// up or down, because the cross product will be between two nearly
		// parallel vectors and starts to approach an undefined state, so we
		// don't draw if the two vectors are less than 1 degree apart
		dot = viewforward.z;	//	same as DotProduct(viewforward, sprup)
								// because sprup is 0, 0, 1
		if ((dot > 0.999848) || (dot < -0.999848))	// cos(1 degree) = 0.999848
			return;

		sr = msin(thing->angles.roll);
		cr = mcos(thing->angles.roll);

		//	CrossProduct(TVec(0, 0, 1), viewforward)
		tvec = Normalize(TVec(viewforward.y, -viewforward.x, 0));
		//	CrossProduct(tvec, TVec(0, 0, 1))
		sprforward = TVec(-tvec.y, tvec.x, 0);
		//	Rotate
		sprright = TVec(tvec.x * cr, tvec.y * cr, tvec.z * cr + sr);
		sprup = TVec(tvec.x * -sr, tvec.y * -sr, tvec.z * -sr + cr);
		break;

	 default:
		I_Error("RenderSprite: Bad sprite type %d", spr_type);
	}

	spritedef_t*	sprdef;
	spriteframe_t*	sprframe;

	// decide which patch to use for sprite relative to player
#ifdef PARANOID
	if ((unsigned)thing->sprite >= MAX_SPRITE_MODELS)
	{
		cond << "Invalid sprite number " << thing->sprite << endl;
		return;
	}
#endif
	sprdef = &sprites[thing->sprite];
#ifdef PARANOID
	if ((thing->frame & FF_FRAMEMASK) >= sprdef->numframes)
	{
		cond << "Invalid sprite frame " << thing->sprite << " : " << thing->frame << endl;
		return;
	}
#endif
	sprframe = &sprdef->spriteframes[thing->frame & FF_FRAMEMASK];

	int			lump;
	boolean		flip;

	if (sprframe->rotate)
	{
		// choose a different rotation based on player view
		//FIXME must use sprforward here?
		angle_t ang = matan(thing->origin.y - vieworg.y, thing->origin.x - vieworg.x);
		dword rot = (ang - thing->angles.yaw + (unsigned)(ANG45 / 2) * 9) >> 29;
		lump = sprframe->lump[rot];
		flip = (boolean)sprframe->flip[rot];
	}
	else
	{
		// use single rotation for all views
		lump = sprframe->lump[0];
		flip = (boolean)sprframe->flip[0];
	}

	if (spritewidth[lump] == -1.0)
	{
		GetSpriteParams(lump);
	}

	draw_vertex_t	sv[4];

	TVec start = -spriteoffset[lump] * sprright;
	TVec end = (spritewidth[lump] - spriteoffset[lump]) * sprright;

	TVec topdelta = (spritetopoffset[lump] - thing->floorclip) * sprup;
	TVec botdelta = (spritetopoffset[lump] - thing->floorclip - spriteheight[lump]) * sprup;

	sv[0] = sprorigin + start + botdelta;
	sv[1] = sprorigin + start + topdelta;
	sv[2] = sprorigin + end + topdelta;
	sv[3] = sprorigin + end + botdelta;

	r_normal = -sprforward;
	r_dist = DotProduct(sprorigin, r_normal);

	if (flip)
	{
		r_saxis = -sprright;
		r_texorg = sv[2];
	}
	else
	{
		r_saxis = sprright;
		r_texorg = sv[1];
	}
	r_taxis = -sprup;

	float lightnum;
	if (thing->frame & FF_FULLBRIGHT || fixedlight)
	{
		lightnum = 255;
	}
	else
	{
		lightnum = sprlight;
	}
	sv[0].l = sv[1].l = sv[2].l = sv[3].l = lightnum;

	if (thing->translucency > 0)
	{
		R_DrawTranslucentPoly(sv, 4, Length(sprorigin - vieworg), lump,
			thing->translucency, thing->translation, true);
	}
	else
	{
		Drawer->DrawSpritePolygon(sv, lump, thing->translucency,
			thing->translation, r_use_fog);
	}
}

//==========================================================================
//
//	R_RenderMobjs
//
//==========================================================================

void R_RenderMobjs(void)
{
	int i;

	if (!r_draw_mobjs)
	{
		return;
	}

    for (i = 0; i < cl_level.numsectors; i++)
    {
		sector_t *sec = &cl_level.sectors[i];

		sprlight = sec->params.lightlevel;//FIXME

	    // Handle all things in sector.
    	for (mobj_base_t *thing = sec->thinglist; thing; thing = thing->snext)
		{
    	    RenderSprite(thing);
		}
    }

	int found;
	do
	{
		found = -1;
		float best_dist = -1;
		for (i = 0; i < MAX_TRANS_SPRITES; i++)
		{
			trans_sprite_t &spr = trans_sprites[i];
			if (!spr.translucency)
			{
				continue;
			}
			if (spr.dist > best_dist)
			{
				found = i;
				best_dist = spr.dist;
			}
		}
		if (found != -1)
		{
			trans_sprite_t &spr = trans_sprites[found];
			r_normal = spr.normal;
			r_dist = spr.pdist;
			r_saxis = spr.saxis;
			r_taxis = spr.taxis;
			r_texorg = spr.texorg;
			if (spr.is_sprite)
			{
				Drawer->DrawSpritePolygon(spr.dv, spr.lump,
					spr.translucency, spr.translation, spr.use_fog);
			}
			else
			{
				Drawer->DrawMaskedPolygon(spr.dv, spr.count, spr.lump,
					spr.translucency, 15, spr.use_fog);
				if (spr.count > 4)
					Z_Free(spr.dv);
			}
			spr.translucency = 0;
		}
	} while (found != -1);

//	memset(trans_sprites, 0, sizeof(trans_sprites));
}

//==========================================================================
//
//	RenderPSprite
//
//==========================================================================

static void RenderPSprite(cl_pspdef_t* psp)
{
	spritedef_t*		sprdef;
	spriteframe_t*		sprframe;
	int					lump;
	boolean				flip;

    // decide which patch to use
#ifdef PARANOID
    if ((unsigned)psp->sprite >= MAX_SPRITE_MODELS)
	{
		con << "R_ProjectSprite: invalid sprite number " << psp->sprite << endl;
		return;
	}
#endif
    sprdef = &sprites[psp->sprite];
#ifdef PARANOID
    if ((psp->frame & FF_FRAMEMASK)  >= sprdef->numframes)
	{
        con << "R_ProjectSprite: invalid sprite frame "
        	<< psp->sprite << " : " << psp->frame << endl;
		return;
	}
#endif
    sprframe = &sprdef->spriteframes[psp->frame & FF_FRAMEMASK];

    lump = sprframe->lump[0];
    flip = (boolean)sprframe->flip[0];

	if (spritewidth[lump] == -1.0)
	{
		GetSpriteParams(lump);
	}

	draw_vertex_t	dv[4];

	TVec sprorigin = vieworg + viewforward;

	float sprx = 160.0 - psp->sx + spriteoffset[lump];
	float spry = 100.0 - psp->sy + spritetopoffset[lump];
	if (viewheight == ScreenHeight)
	{
		spry -= cl.pspriteSY;
	}

	//	1 / 160 = 0.00625
	TVec start = sprorigin - (sprx * 0.00625) * viewright;
	TVec end = start + (spritewidth[lump] * 0.00625) * viewright;

	//	1 / 160.0 * 120 / 100 =	0.0075
	TVec topdelta = (spry * 0.0075) * viewup;
	TVec botdelta = topdelta - (spriteheight[lump] * 0.0075) * viewup;

	dv[0] = start + botdelta;
	dv[1] = start + topdelta;
	dv[2] = end + topdelta;
	dv[3] = end + botdelta;

	r_normal = -viewforward;
	r_dist = DotProduct(dv[0], r_normal);

	if (flip)
	{
		r_saxis = -(viewright * 160);
		r_texorg = dv[2];
	}
	else
	{
		r_saxis = viewright * 160;
		r_texorg = dv[1];
	}
	r_taxis = -(viewup * 100 * 4 / 3);

	float lightnum;
	if (psp->frame & FF_FULLBRIGHT)
	{
		lightnum = 255;
	}
	else
	{
		lightnum = sprlight;
	}
	dv[0].l = dv[1].l = dv[2].l = dv[3].l = lightnum;

	Drawer->DrawSpritePolygon(dv, lump, cl.translucency, 0, r_use_fog);
}

//==========================================================================
//
// 	R_DrawPlayerSprites
//
//==========================================================================

void R_DrawPlayerSprites(void)
{
    int         i;
    cl_pspdef_t	*psp;

	if (!r_draw_psprites || r_chasecam)
	{
		return;
	}

    // get light level
	sprlight = r_viewleaf->sector->params.lightlevel;

    // add all active psprites
	for (i = 0, psp = cl.psprites; i < NUMPSPRITES; i++, psp++)
	{
		if (psp->sprite != -1)
		{
			RenderPSprite(psp);
		}
	}
}

//==========================================================================
//
//	R_DrawCroshair
//
//==========================================================================

void R_DrawCroshair(void)
{
	if (croshair)
	{
		if (croshair_trans < 0)		croshair_trans = 0;
		if (croshair_trans > 100)	croshair_trans = 100;

		int			cy;
		if (screenblocks < 11)
			cy = (200 - sb_height) / 2;
		else
			cy = 100;
		int handle = R_RegisterPic(va("CROSHAI%i", (int)croshair), PIC_PATCH);
		R_DrawPic(160, cy, handle, croshair_trans);
	}
}

//==========================================================================
//
//  R_DrawSpritePatch
//
//==========================================================================

void R_DrawSpritePatch(int x, int y, int sprite, int frame, int rot, int translation)
{
    spritedef_t*	sprdef;
    spriteframe_t*	sprframe;
    boolean			flip;
	int				lump;

    sprdef = &sprites[sprite];
    sprframe = &sprdef->spriteframes[frame & FF_FRAMEMASK];
    flip = (boolean)sprframe->flip[rot];
	lump = sprframe->lump[rot];
	Drawer->DrawSpriteLump(x, y, lump, translation, flip);
}

